#ifndef __JABYENGINE_PRIMITIVE_SUPPORT_TYPES_HPP__
#define __JABYENGINE_PRIMITIVE_SUPPORT_TYPES_HPP__
#include "../gpu_types.hpp"
#include "../../System/IOPorts/gpu_io.hpp"

namespace JabyEngine {
    namespace GPU {
        namespace internal {
            template<typename T>
            struct CodeBase {
                static constexpr auto BitCorrection = 24;
                static constexpr auto CmdID         = BitRange::from_to(29 - BitCorrection, 31 - BitCorrection);

                uint8_t value = bit::value::set_normalized(0u, CmdID.with(T::CmdValue));

                constexpr CodeBase() = default;
                constexpr CodeBase(const T& code) : value(code.value) {
                }

                constexpr T& set(Bit bit) {
                    this->value = bit::set(this->value, bit);
                    return static_cast<T&>(*this);
                }

                constexpr T& set(ClearBit bit) {
                    this->value = bit::set(this->value, bit);
                    return static_cast<T&>(*this);
                }
            };

            // Concept for now
            template<typename T>
            struct Hooked {
                uint32_t hook;
                T        primitive;
            };

            template<typename T>
            struct is_render_primitive {
                static constexpr bool value = false;
            };

            #define __jaby_engine_declare_render_primitive(type) \
            template<> \
            struct is_render_primitive<type> { \
                static constexpr bool value = true; \
            }
        }

        // Reexport for easier use
        typedef JabyEngine::GPU_IO::SemiTransparency SemiTransparency; 
        typedef JabyEngine::GPU_IO::TexturePageColor TexturePageColor;

        struct PageClut {
            uint16_t value = 0;
            
            constexpr PageClut() = default;
            constexpr PageClut(uint16_t x, uint16_t y) : value((y << 6) | ((x >> 4) & 0x3f)) {
            }
        };

        struct TPage {
            static constexpr auto TexturePageX     = BitRange::from_to(0, 3);
            static constexpr auto TexturePageY     = BitRange::from_to(4, 4);
            static constexpr auto SemiTransparency = BitRange::from_to(5, 6);
            static constexpr auto TextureClut      = BitRange::from_to(7, 8);

            uint16_t value = 0;

            constexpr TPage() = default;
            constexpr TPage(uint16_t x, uint16_t y, ::JabyEngine::GPU::SemiTransparency transparency, TexturePageColor clut_color) : value(TexturePageX.as_value(x >> 6) | TexturePageY.as_value(y >> 8) | SemiTransparency.as_value(static_cast<uint16_t>(transparency)) | TextureClut.as_value(static_cast<uint16_t>(clut_color))) {
            }
        };
    }
}

#endif //!__JABYENGINE_PRIMITIVE_SUPPORT_TYPES_HPP__